/**
 * The mergeable-store module
 * @packageDocumentation
 * @module mergeable-store
 * @since v5.0.0
 */

import {CellOrUndefined, Content, Store, ValueOrUndefined} from './store.d.ts';
import {Id, Ids} from './common';

/**
 * The Hash type
 * @category Stamps
 */
export type Hash = number;

/**
 * The Time type
 * @category Stamps
 */
export type Time = string;

/**
 * The Stamp type
 * @category Stamps
 */
export type Stamp<Thing, Hashed extends boolean = false> = Hashed extends true
  ? [time: Time, thing: Thing, hash: Hash]
  : [time: Time, thing: Thing];

// ContentHashes
export type ContentHashes = [time: Time, [tablesHash: Hash, valuesHash: Hash]];

// TablesStamp
export type TablesStamp<Hashed extends boolean = false> = Stamp<
  {[tableId: Id]: TableStamp<Hashed>},
  Hashed
>;

// TableHashes
export type TableHashes = {[tableId: Id]: Hash};

// TableIdsDiff
export type TableIdsDiff = Ids;

// TableStamp
export type TableStamp<Hashed extends boolean = false> = Stamp<
  {[rowId: Id]: RowStamp<Hashed>},
  Hashed
>;

// RowHashes
export type RowHashes = {[tableId: Id]: {[rowId: Id]: Hash}};

// RowIdsDiff
export type RowIdsDiff = {[tableId: Id]: Ids};

// RowStamp
export type RowStamp<Hashed extends boolean = false> = Stamp<
  {[cellId: Id]: CellStamp<Hashed>},
  Hashed
>;

// CellHashes
export type CellHashes = {[tableId: Id]: {[rowId: Id]: {[cellId: Id]: Hash}}};

// CellStamp
export type CellStamp<Hashed extends boolean = false> = Stamp<
  CellOrUndefined,
  Hashed
>;

// ValuesStamp
export type ValuesStamp<Hashed extends boolean = false> = Stamp<
  {[valueId: Id]: ValueStamp<Hashed>},
  Hashed
>;

// ValuesHashes
export type ValuesHashes = {[valueId: Id]: Hash};

// ValueStamp
export type ValueStamp<Hashed extends boolean = false> = Stamp<
  ValueOrUndefined,
  Hashed
>;

/**
 * The MergeableContent type
 * @category Mergeable
 */
export type MergeableContent = Stamp<
  [mergeableTables: TablesStamp<true>, mergeableValues: ValuesStamp<true>],
  true
>;

/**
 * The MergeableChanges type
 * @category Mergeable
 */
export type MergeableChanges = Stamp<
  [mergeableTables: TablesStamp, mergeableValues: ValuesStamp, isChanges: 1]
>;

/**
 * The MergeableStore type
 * @category Mergeable
 */
export interface MergeableStore extends Store {
  //
  /**
   * The getId method
   * @category Getter
   */
  getId(): Id;

  /**
   * The getMergeableContent method
   * @category Getter
   */
  getMergeableContent(): MergeableContent;

  /**
   * The getMergeableContentHashes method
   * @category Syncing
   */
  getMergeableContentHashes(): ContentHashes;

  /**
   * The getMergeableTableHashes method
   * @category Syncing
   */
  getMergeableTableHashes(): TableHashes;

  /**
   * The getMergeableTableIdsDiff method
   * @category Syncing
   */
  getMergeableTableIdsDiff(relativeTo: TableHashes): TableIdsDiff;

  /**
   * The getMergeableRowHashes method
   * @category Syncing
   */
  getMergeableRowHashes(tablesDelta: TableIdsDiff): RowHashes;

  /**
   * The getMergeableRowIdsDiff method
   * @category Syncing
   */
  getMergeableRowIdsDiff(relativeTo: RowHashes): RowIdsDiff;

  /**
   * The getMergeableCellHashes method
   * @category Syncing
   */
  getMergeableCellHashes(tableDelta: RowIdsDiff): CellHashes;

  /**
   * The getMergeableTablesChanges method
   * @category Syncing
   */
  getMergeableTablesChanges(relativeTo: CellHashes): TablesStamp;

  /**
   * The getMergeableValuesHashes method
   * @category Syncing
   */
  getMergeableValuesHashes(): ValuesHashes;

  /**
   * The getMergeableValuesChanges method
   * @category Syncing
   */
  getMergeableValuesChanges(relativeTo: ValuesHashes): ValuesStamp;

  /**
   * The setMergeableContent method
   * @category Setter
   */
  setMergeableContent(mergeableContent: MergeableContent): MergeableStore;

  /**
   * The setDefaultContent method
   * @category Setter
   */
  setDefaultContent(content: Content): MergeableStore;

  /**
   * The getTransactionMergeableChanges method
   * @category Getter
   */
  getTransactionMergeableChanges(): MergeableChanges;

  /**
   * The applyMergeableChanges method
   * @category Setter
   */
  applyMergeableChanges(
    mergeableChanges: MergeableChanges | MergeableContent,
  ): MergeableStore;

  /**
   * The merge method
   * @category Setter
   */
  merge(mergeableStore: MergeableStore): MergeableStore;
}

/**
 * The createMergeableStore function
 * @category Creation
 */
export function createMergeableStore(id: Id): MergeableStore;
